package com.restlet.client.async;

import com.restlet.client.function.Consumer;

import java.util.List;

/**
 * An abstraction to work with the collections of values where each values is resolved
 * This is more efficient then simple promise of list, because it requires less memory,
 * and since streams can be closed
 *
 * @param <T> type of values to be passed into stream
 */
public interface AsyncStream<T> {

    /**
     * Each time a value appears in a stream consumer consume method will be closed once
     * Consumers can close the stream
     * If consumer is added to closed stream, they will receive all values only
     * if there were no other consumers in a stream
     * If consumer throws an exception during consume() method called,
     *
     * @param consumer consumer
     * @return reference to the same stream
     */
    AsyncStream<T> forEach(Consumer<T> consumer);

    /**
     * Returns items which have been put in a stream since invocation of this method
     * till stream is closed
     *
     * @return promise with values, which is resolved when stream is closed
     */
    Promise<List<T>> asList();

    /**
     * Closes the stream, can be called multiple times,
     * but close promise will be resolved only ones
     * After stream is closed no new values should be added in a stream
     * Method is safe to call multiple times
     */
    AsyncStream<T> closeIfStillOpened();

    /**
     * Closes the stream with error. Close promise will be rejected
     *
     * @param e error which caused stream closure
     */
    void closeIfStillOpened(Throwable e);

    /**
     * Closes the stream when provided promise is resolved
     *
     * @param closePromise promise to watch
     */
    void closeWhen(Promise<Void> closePromise);

    /**
     * Checks if the stream is currently closed
     *
     * @return stream closed state
     */
    boolean isClosed();

    /**
     * Promise for stream closing,
     * It is resolved when close() method is called on stream
     *
     * @return promise for stream closing
     */
    Promise<Void> closePromise();
}
