package com.restlet.client.async;

import com.restlet.client.utils.NotNull;
import com.restlet.client.utils.Nullable;

/**
 * A callback which allows to track state
 *
 * @param <T>
 */
public interface Deferred<T> {
    /**
     * Returns resolved value. Note that deferred can be resolved with null value
     *
     * @return resolved value
     */
    T getResult();

    /**
     * Returns rejection exception
     *
     * @return a rejection exception
     */
    Throwable getError();

    /**
     * Returns current state of deferred
     *
     * @return state of deferred
     */
    @NotNull
    State getState();

    /**
     * Returns a promise associated with this deferred
     *
     * @return a promise
     */
    Promise<T> promise();

    /**
     * Resolves a promise in nearby future with specified (any code running in current thread should be allowed to finish before value is resolved)
     * Will throw an error if deferred isn't in #State.PENDING
     *
     * @param result value to resolve with
     */
    void resolveLater(T result);

    /**
     * Instantly resolves promise with null
     * Will throw an error if deferred isn't in #State.PENDING
     */
    void resolve();

    /**
     * Instantly resolves a deferred with specified value
     * Will throw an error if deferred isn't in #State.PENDING
     *
     * @param result a value to be resolved with
     */
    void resolve(@Nullable T result);

    /**
     * Instantly rejects a promise with specified error
     * Will throw an error if deferred isn't in #State.PENDING
     *
     * @param error error to reject with
     */
    void reject(Throwable error);

    /**
     * Rejects a promise and logs the given message at level warning.
     */
    void reject(String warningMessage);

    /**
     * Resolver for easier linking of deferred objects to promises
     *
     * @return a handler which will resolve this deferred with supplied value once it is called
     */
    PromiseHandler<T> resolver();

    /**
     * Rejector for easier linking of deferred objects to promise
     *
     * @return a handler which will reject this deferred with supplied error, once it is called
     */
    PromiseHandler<Throwable> rejector();

    /**
     * Adds a new handler to the deferred. Should never be called outside of Promise based classes.
     * Will throw an error if deferred isn't in #State.PENDING
     *
     * @param handler a handler which will be called once deferred changes it's state
     */
    void addDeferredResolvedHandler(DeferredResolvedHandler handler);

    /**
     * Resolves deferred with null. Never throws any errors
     *
     * @return true if state of deferred were changed, false otherwise
     */
    boolean safeResolve();

    /**
     * Resolves deferred with null. Never throws any errors
     *
     * @param value value to resolve with
     * @return true if state of deferred were changed, false otherwise
     */
    boolean safeResolve(T value);

    /**
     * Rejects deferred with supplied error. Never throws any errors
     *
     * @param error error to reject with
     * @return true if state of deferred were changed, false otherwise
     */
    boolean safeReject(Throwable error);

    enum State {
        PENDING, SUCCESS, ERROR
    }

    interface DeferredResolvedHandler {
        void onResolve(Deferred<?> deferred);
    }
}
