package com.restlet.client.async;

import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.utils.Holder;

/**
 * Promises implementation,compatible with Promise/A+ proposal
 *
 * @param <C>
 */
public interface Promise<C> {

    /**
     * Same as {@link #then(PromiseHandler, PromiseHandler)} , but allows to specify only success handler
     *
     * @deprecated Use {@link Promise#map(Function)} instead
     */
    @Deprecated
    <T> Promise<T> then(PromiseHandler successHandler);

    /**
     * Legacy method from pre java8 day, essentially a mix of flatMap and map.
     * This behaves exactly the same way as a then method from Promise/A+ proposal
     * Promise handlers can return either a concrete value or a promise of value.
     * When promise is returned from handler, it is automatically unwrapped.
     * If any of handlers throws an exception, then returned promise will be rejected
     * At least one the handlers should be not null, otherwise an error is thrown
     *
     * @param successHandler called when this promise is resolved, returned value will be used to resolve returned promise
     * @param errorHandler   called when this promise is rejected, returned value will be used to resolve returned promise. This handler is NOT invoked if success handler throws an error
     * @param <T>            type of value in returned promise
     * @deprecated Use {@link Promise#map(Function)} instead
     */
    @Deprecated
    <T> Promise<T> then(final PromiseHandler successHandler, final PromiseHandler<Throwable> errorHandler);

    /**
     * @deprecated use {@link #recoverOnReject(PromiseHandler)} instead
     */
    @Deprecated
    <T> Promise<T> onError(PromiseHandler<Throwable> errorHandler);

    /**
     * Calls supplied handler with error, if promise was rejected
     * If promise is rejected, then handler return value will be used to obtain new value which will be used to obtain new promise
     * Promise values returned from error handler will be automatically unwrapped
     *
     * @param errorHandler error recovery. If this promise was rejected, return value from promise handler will be used for recovery.
     *                     If handler itself throws an error, then returned promise will be rejected as well.
     * @return original promise if this promise was resolved, otherwise recovered promise from errorHandler
     */
    Promise<C> recoverOnReject(PromiseHandler<Throwable> errorHandler);

    /**
     * When this promise is rejected, returned promise is resolved/rejected when parameter promise is resolved/rejected
     * When this promise is resolved, returned promise is resolved with the same value as this promise
     *
     * @param promise a promise to be used when this promise is rejected
     */
    Promise<C> recoverOnReject(Promise<C> promise);

    /**
     * When this promise is rejected, returned promise will resolved with supplied parameter value
     * When this promise is resolved, returned promise will resolved with same value as this promise
     *
     * @param value value to be used when this promise is rejected
     */
    Promise<C> recoverOnReject(C value);

    /**
     * Defaults recover on reject which will send back null
     *
     * @return
     */
    Promise<C> recoverOnReject();

    <T> Promise<T> flatRecoverOnReject(final Function<Throwable, Promise<T>> handler);

    <T> Promise<T> cast();

    boolean isDone();

    /**
     * Maps resolved value of this promise to a promise returned from supplied handler.
     *
     * @param handler function which maps resolved value of this promise to another promises
     * @param <T>     type of value of promise returned from handler
     * @return a promise, which will be resolved/rejected with the value of promise returned with the handler, assuming current promise was resolved and handler itself didn't threw an error.
     */
    <T> Promise<T> flatMap(final Function<C, Promise<T>> handler);

    /**
     * Maps resolved value of this promise to a promise returned from supplied handler.
     *
     * @param handler function which maps resolved value of this promise to another promises
     * @param <T>     type of value of promise returned from handler
     * @return a promise, which will be resolved/rejected with the value of promise returned with the handler, assuming current promise was resolved and handler itself didn't threw an error.
     */
    <T> Promise<T> flatMapIfNotNull(final Function<C, Promise<T>> handler);

    /**
     * When this promise is resolved, will return a promise which will be resolved/rejected when provided promise is resolved/rejected
     *
     * @param promise promise to map to.
     * @param <T>     type of value in promise
     * @return when this promise is resolved, returns a promise which is resolved/rejected when parameter promise is resolved/rejected
     */
    <T> Promise<T> flatMap(final Promise<T> promise);

    /**
     * Applies the function to the resolved value of the promise, and returns a promise of this new value
     * if this promise is resolved.
     *
     * @param handler a function which maps current value of promise to another value.
     * @param <T>     type of the value returned from the handler
     * @return a new promise, which will be either resolved with value of supplied handler, or rejected if the handler throws an error or this promise itself was rejected
     */
    <T> Promise<T> map(final Function<C, T> handler);

    /**
     * Applies the function to the resolved value of the promise, and returns a promise of this new value
     * if this promise is resolved.
     *
     * @param handler a function which maps current value of promise to another value.
     * @param <T>     type of the value returned from the handler
     * @return a new promise, which will be either resolved with value of supplied handler, or rejected if the handler throws an error or this promise itself was rejected
     */
    <T> Promise<T> mapIfNotNull(final Function<C, T> handler);

    /**
     * Calls a supplied consumer with error, when promise is rejected
     *
     * @param consumer consumer which will be notified of error. If consumer throws an error,
     *                 it will be caught an logged into Stdout or default output log
     * @return original promise on which this method is called
     */
    Promise<C> doOnReject(Consumer<Throwable> consumer);

    /**
     * Calls a supplied consumer with a value, when promise was resolved
     *
     * @param consumer consumer which will be notified of value. If consumer throws an error,
     *                 it will be caught an logged into Stdout or default output log
     * @return original promise on which this method is called
     */
    Promise<C> doOnResolve(Consumer<C> consumer);

    /**
     * Same behaviour as the doOnResolve but the consumer is triggered only if the result of the promise is not null
     *
     * @param consumer consumer which will be notified of value if the value is not null. If consumer throws an error,
     *                 it will be caught an logged into Stdout or default output log
     * @return original promise on which this method is called
     */
    Promise<C> doOnResolveIfNotNull(Consumer<C> consumer);

    /**
     * Calls a supplied consumer when this promise is either resolved&reject
     *
     * @param consumer consumer which will be notified of completion.
     *                 First parameter is a resolved value (can be null), second is an error
     *                 If consumer throws an error,
     *                 it will be caught an logged into Stdout or default output log
     * @return original promise
     */
    Promise<C> doFinally(Consumer<Deferred> consumer);

    /**
     * @return Returns a deferred object representing this promise
     */
    Deferred getDeferred();

    Promise<Void> castToVoidPromise();

    Promise<C> assignTo(Holder<C> parentValueHolder);
}
