package com.restlet.client.async;

import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Executor;
import com.restlet.client.function.Function;
import com.restlet.client.utils.Holder;
import com.restlet.client.utils.Nullable;
import com.restlet.client.utils.Sequence;

import java.util.Collections;
import java.util.List;

import static java.util.Arrays.asList;
import static com.restlet.client.utils.ListUtils.newFilledList;
import static com.restlet.client.utils.Objects.isNullOrEmpty;

public abstract class Promises {

    protected Promises() {
    }

    /**
     * @deprecated Use {@link Promises#of(Object) instead}
     */
    @SuppressWarnings("unchecked")
    @Deprecated
    public static <F> Promise<F> when(Object promiseOrValue) {
        if (promiseOrValue instanceof Promise) {
            return (Promise<F>) promiseOrValue;
        }
        Deferred<F> deferred = new DeferredImpl<>();
        deferred.resolve((F) promiseOrValue);
        return deferred.promise();
    }

    /**
     * Creates a promise, which is resolved when all of supplied values are resolved successfully,
     * Otherwise result promise is rejected
     *
     * @param promises promise list
     */
    public static <F> Promise<List<F>> all(final List<Promise<F>> promises) {
        if (isNullOrEmpty(promises)) {
            return Promises.of(Collections.<F>emptyList());
        }

        final Holder<Integer> counter = Holder.of(0);
        final Deferred<List<F>> deferred = new DeferredImpl<>();
        final List<F> values = newFilledList(promises.size(), null);

        Sequence.of(promises)
                .each(new BiConsumer<Promise<F>, Integer>() {
                    @Override
                    public void consume(Promise<F> promise, final Integer index) {
                        promise
                                .doOnResolve(new Consumer<F>() {
                                    @Override
                                    public void consume(F result) {
                                        values.set(index, result);
                                        if (deferred.getState() == Deferred.State.PENDING) {
                                            counter.value++;
                                            if (promises.size() == counter.value) {
                                                deferred.resolve(values);
                                            }
                                        }
                                    }
                                })
                                .doOnReject(new Consumer<Throwable>() {
                                    @Nullable
                                    @Override
                                    public void consume(@Nullable final Throwable error) {
                                        if (deferred.getState() == Deferred.State.PENDING) {
                                            deferred.reject(error);
                                        }
                                    }
                                });
                    }
                });
        return deferred.promise();
    }

    /**
     * Creates a rejected promise with specified error
     *
     * @param e   error
     * @param <T> type of promise
     * @return a promise rejected with supplied error
     */
    public static <T> Promise<T> reject(final Throwable e) {
        Deferred<T> deferred = new DeferredImpl<>();
        deferred.reject(e);
        return deferred.promise();
    }

    /**
     * Creates a rejected promise with specified message. The message will be logged at level <code>warning</code>.
     */
    public static <T> Promise<T> reject(final String warningMessage) {
        Deferred<T> deferred = new DeferredImpl<>();
        deferred.reject(warningMessage);
        return deferred.promise();
    }

    /**
     * Calls a callback with the values of both promises, once both promises are resolved
     *
     * @param firstPromise  first promise
     * @param secondPromise second promise
     * @param callback      a callback to be called once both promises are resolved. If callback throws an error once it is invoked, returned promise will be rejected
     * @param <F>           type of value in first promise
     * @param <S>           type of value in second promise
     * @param <R>           resulting value returned by callback
     * @return returns a promise which is resolved once both supplied promises are resolved, with the value returned from callback. If any of supplied promises are rejected or callback throws an error once executed, returned promise will be rejected.
     */
    @SuppressWarnings("unchecked")
    public static <F, S, R> Promise<R> combine(final Promise<F> firstPromise, final Promise<S> secondPromise, final CombineCallback<F, S, R> callback) {
        final Object[] vals = new Object[2];
        List<Promise<Void>> safePromises = asList(
                firstPromise
                        .doOnResolve(new Consumer<F>() {
                            @Override
                            public void consume(@Nullable final F value) {
                                vals[0] = value;
                            }
                        })
                        .castToVoidPromise(),
                secondPromise
                        .doOnResolve(new Consumer<S>() {
                            @Override
                            public void consume(@Nullable final S value) {
                                vals[1] = value;
                            }
                        })
                        .castToVoidPromise());

        return Promises.all(safePromises)
                .map(new Function<List<Void>, R>() {
                    @Nullable
                    @Override
                    public R apply(@Nullable final List<Void> input) {
                        return callback.on((F) vals[0], (S) vals[1]);
                    }
                });
    }

    @SuppressWarnings("unchecked")
    public static <T> T get(Promise<T> promise) {
        final Holder<T> result = Holder.empty();
        final Holder<Throwable> error = Holder.empty();
        if (!promise.isDone()) {
            throw new IllegalArgumentException("Promise should be in done state");
        }
        promise.doOnReject(new Consumer<Throwable>() {
            @Override
            public void consume(@Nullable final Throwable e) {
                error.value = e;
            }
        });
        if (error.hasValue()) {
            throw new RuntimeException(error.value);
        }
        promise.doOnResolve(new Consumer<T>() {
            @Override
            public void consume(@Nullable final T value) {
                result.value = value;
            }
        });
        return result.value;
    }

    /**
     * Safer way to create a promise. Automatically creates deferred object and passes it to the supplied consumer.
     * If consumer throws an exception, then whole promise is rejected with this value
     *
     * @param consumer takes a deferred and resolves/rejects it later. If consumer throws an error, deferred will be rejected automatically
     * @param <T>      type of value of return promise
     * @return a newly created promise
     */
    public static <T> Promise<T> create(Consumer<Deferred<T>> consumer) {
        DeferredImpl<T> deferred = new DeferredImpl<>();
        try {
            consumer.consume(deferred);
        } catch (Throwable e) {
            if (deferred.getState() == Deferred.State.PENDING) {
                deferred.reject(e);
            } else {
                DeferredImpl.LOG_SERVICE.error("Promise creator threw an error after resolving/rejecting the promise", e);
            }
        }
        return deferred.promise();
    }

    public static <T> Promise<T> create(final Executor executor) {
        try {
            executor.execute();
        } catch (Throwable e) {
            return reject(e);
        }
        return of();
    }

    public interface CombineCallback<F, S, R> {
        R on(F first, S second);

    }

    /**
     * Creates a promise from supplied value
     * This is different from {@link #when(Object)} method, since it automatically inherits type from supplied value, therefore it is more typesafe
     *
     * @param value a value
     * @param <T>   type of value
     * @return a promise of supplied value
     */
    public static <T> Promise<T> of(final T value) {
        Deferred<T> deferred = new DeferredImpl<>();
        deferred.resolve(value);
        return deferred.promise();
    }

    public static <T> Promise<T> of() {
        return Promises.of(null);
    }

    public static Promise<Void> voidPromise() {
        return Promises.of();
    }
}
