package com.restlet.client.async.impl;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.log.LogService;
import com.restlet.client.log.impl.MockLogService;
import com.restlet.client.utils.NotNull;
import com.restlet.client.utils.Nullable;

import java.util.ArrayList;
import java.util.List;

import static com.restlet.client.async.Deferred.State.ERROR;
import static com.restlet.client.utils.Objects.isNullOrEmpty;

/**
 * A callback which allows to track state
 *
 * @param <T>
 */
public class DeferredImpl<T> implements Deferred<T> {
    public static LogService LOG_SERVICE = new MockLogService(DeferredImpl.class);

    private T result;
    private Throwable error;
    @Nullable
    private List<DeferredResolvedHandler> handlers;

    private State state = State.PENDING;

    public T getResult() {
        return result;
    }

    public Throwable getError() {
        return error;
    }

    public State getState() {
        return state;
    }

    @Override
    public void addDeferredResolvedHandler(@NotNull DeferredResolvedHandler handler) {
        if (getState() != State.PENDING) {
            throw new IllegalStateException("Deferred has been already resolved");
        }
        if (handlers == null) {
            handlers = new ArrayList<>();
        }
        handlers.add(handler);
    }

    @Override
    public boolean safeResolve() {
        return safeResolve(null);
    }

    @Override
    public boolean safeResolve(final T value) {
        if (state == State.PENDING) {
            resolve(value);
            return true;
        }
        return false;
    }

    @Override
    public boolean safeReject(final Throwable error) {
        if (state == State.PENDING) {
            reject(error);
            return true;
        }
        return false;
    }

    public Promise<T> promise() {
        return new PromiseImpl<>(this);
    }

    public void resolveLater(T result) {
        resolve(result);
    }

    public void resolve() {
        resolve(null);
    }

    public void resolve(T result) {
        complete(State.SUCCESS, result, null);
    }

    public void reject(Throwable error) {
        if (error != null) {
            LOG_SERVICE.error(error.getMessage(), error);
        }

        complete(ERROR, null, error);
    }

    public void reject(final String warningMessage) {
        if (!isNullOrEmpty(warningMessage)) {
            LOG_SERVICE.warning(warningMessage);
        }

        complete(ERROR, null, null);
    }

    protected void complete(@NotNull State state, @Nullable T result, @Nullable Throwable error) {

        if (this.state != State.PENDING) {
            throw new IllegalStateException("This promise has been already resolved");
        }

        this.error = error;
        this.result = result;
        this.state = state;

        if (handlers == null) {
            return;
        }

        for (DeferredResolvedHandler handler : handlers) {

            try {
                handler.onResolve(this);

            } catch (Exception e) {
                // do nothing
            }
        }
        handlers = null;
    }

    public PromiseHandler<T> resolver() {
        return new PromiseHandler<T>() {
            @Override
            public Object on(final T error) {
                resolve(error);
                return null;
            }
        };
    }

    public PromiseHandler<Throwable> rejector() {
        return new PromiseHandler<Throwable>() {
            @Override
            public Object on(final Throwable error) {
                reject(error);
                return null;
            }
        };
    }
}
