package com.restlet.client.async.impl;

import com.restlet.client.async.AsyncStream;
import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.function.Function;
import com.restlet.client.function.Consumer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

public class WritableStream<T> implements AsyncStream<T> {

    private List<T> stack = new ArrayList<>();
    private List<Consumer<T>> consumers = new ArrayList<>();
    private final Deferred<Void> closeDeferred;

    public WritableStream() {
        this(Collections.<T>emptyList());
    }

    public WritableStream(List<T> initialValue) {
        stack.addAll(initialValue);
        closeDeferred = new DeferredImpl<>();
    }

    public void accept(List<T> list) throws IllegalStateException {
        for (T val : list) {
            accept(val);
        }
    }

    public void accept(T value) throws IllegalStateException {
        if (isClosed()) {
            throw new IllegalStateException("Stream is closed");
        }
        if (consumers.isEmpty()) {
            stack.add(value);
        } else {
            for (Consumer<T> consumer : consumers) {
                consume(consumer, value);
            }
        }
    }

    @Override
    public AsyncStream<T> forEach(final Consumer<T> consumer) {
        if (!isClosed()) {
            consumers.add(consumer);
        }
        if (stack.isEmpty()) {
            return this;
        }
        final Iterator<T> iterator = stack.iterator();
        while (iterator.hasNext()) {
            final T next = iterator.next();
            iterator.remove();
            consume(consumer, next);
        }
        return this;
    }

    private void consume(Consumer<T> consumer, T value) {
        try {
            consumer.consume(value);
        } catch (Exception e) {
            //FIXME GWT.reportUncaughtException(new IllegalStateException("Consumer thrown an error when consuming value: " + value,e));
        }
    }

    @Override
    public Promise<List<T>> asList() {
        final List<T> returnValue = new ArrayList<>();
        if (isClosed()) {
            return Promises.of(returnValue);
        }
        forEach(new Consumer<T>() {
            @Override
            public void consume(final T t) {
                returnValue.add(t);
            }
        });

        return closeDeferred.promise().map(new Function<Void, List<T>>() {
            @Override
            public List<T> apply(Void whatever) {
                return returnValue;
            }
        });
    }

    @Override
    public WritableStream<T> closeIfStillOpened() {
        if (!isClosed()) {
            closeDeferred.resolve();
        }

        return this;
    }

    @Override
    public void closeIfStillOpened(final Throwable e) {
        if (!isClosed()) {
            closeDeferred.reject(e);
        }
    }

    @Override
    public void closeWhen(final Promise<Void> closePromise) {
        closePromise
                .doOnResolve(new Consumer<Void>() {
                    @Override
                    public void consume(Void whatever) {
                        closeIfStillOpened();
                    }
                })
                .doOnReject(new Consumer<Throwable>() {
                    @Override
                    public void consume(Throwable throwable) {
                        closeIfStillOpened(throwable);
                    }
                });
    }

    @Override
    public boolean isClosed() {
        return closeDeferred.getState() != Deferred.State.PENDING;
    }

    @Override
    public Promise<Void> closePromise() {
        return closeDeferred.promise();
    }

}
