package com.restlet.client.model;

import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.utils.Maybe;
import com.restlet.client.utils.Sequence;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import static com.restlet.client.model.UpdateType.None;
import static com.restlet.client.utils.ListUtils.first;

public class EntityTreeNode {

    public final EntityTo entity;
    public final List<EntityTreeNode> children = new ArrayList<>();
    public UpdateType updateType;
    public UpdateStatus updateStatus;

    public EntityTreeNode(EntityTo entity) {
        this.entity = entity;
        this.updateType = None;
    }

    public EntityTreeNode(EntityTo entity, UpdateType updateType) {
        this(entity);
        this.updateType = updateType;
    }

    public EntityTreeNode setUpdateType(UpdateType updateType) {
        this.updateType = updateType;
        return this;
    }

    public EntityTreeNode setUpdateStatus(UpdateStatus updateStatus) {
        this.updateStatus = updateStatus;
        return this;
    }

    public EntityTreeNode addChild(final EntityTreeNode child) {
        children.add(child);
        return this;
    }

    public EntityTreeNode addChildren(final Collection<EntityTreeNode> children) {
        this.children.addAll(children);
        return this;
    }

    @Override
    public String toString() {
        return "ProjectTreeNode{" +
                "entity=" + entity +
                ", children=" + children +
                ", updateType=" + updateType +
                '}';
    }

    public EntityTreeNode retrieveChildByName(String entityName) {
        for (EntityTreeNode child : children) {
            if (entityName.equals(child.entity.getName())) {
                return child;
            }
        }
        throw new IllegalArgumentException("Child named '" + entityName + "' not found.");
    }

    public EntityTreeNode retrieveChildById(String id) {
        for (EntityTreeNode child : children) {
            if (id.equals(child.entity.getId())) {
                return child;
            }
        }
        throw new IllegalArgumentException("Child with id '" + id + "' not found.");
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        EntityTreeNode other = (EntityTreeNode) o;

        return entity.equals(other.entity)
                && children.equals(other.children)
                && updateType == other.updateType;

    }

    @Override
    public int hashCode() {
        int result = entity.hashCode();
        result = 31 * result + children.hashCode();
        result = 31 * result + (updateType != null ? updateType.hashCode() : 0);
        return result;
    }

    public List<EntityTo> flattenEntitiesRecursively() {
        return reduce(new Function<EntityTreeNode, Maybe<EntityTo>>() {
            @Override
            public Maybe<EntityTo> apply(final EntityTreeNode node) {
                return Maybe.of(node.entity);
            }
        });
    }

    public List<EntityTo> retrieveAllChildrenRecursively() {
        return Sequence.of(children)
                .flatMap(new Function<EntityTreeNode, Iterable<EntityTo>>() {
                    @Override
                    public Iterable<EntityTo> apply(EntityTreeNode input) {
                        return input.flattenEntitiesRecursively();
                    }
                })
                .toList();
    }

    public Maybe<EntityTreeNode> find(final Function<EntityTreeNode, Boolean> filter) {
        final List<EntityTreeNode> candidates = new ArrayList<>();
        if (filter.apply(this)) {
            candidates.add(this);
        }

        Sequence.of(children)
                .omitNulls()
                .each(new BiConsumer<EntityTreeNode, Integer>() {
                    @Override
                    public void consume(EntityTreeNode child, Integer index) {
                        child.find(filter)
                                .ifPresent(new Consumer<EntityTreeNode>() {
                                    @Override
                                    public void consume(final EntityTreeNode foundCandidate) {
                                        candidates.add(foundCandidate);
                                    }
                                });
                    }
                });

        if (candidates.isEmpty()) {
            return Maybe.empty();

        } else if (candidates.size() == 1) {
            return Maybe.of(first(candidates));
        }

        throw new IllegalArgumentException("Cannot find a node corresponding to the search.");
    }

    public void walk(final Consumer<EntityTreeNode> visitor) {
        visitor.consume(this);
        Sequence.of(children).each(new BiConsumer<EntityTreeNode, Integer>() {
            @Override
            public void consume(EntityTreeNode child, Integer index) {
                child.walk(visitor);
            }
        });
    }

    public EntityTreeNode recursiveSetUpdateType(final UpdateType newUpdateType) {
        this.walk(new Consumer<EntityTreeNode>() {
            @Override
            public void consume(final EntityTreeNode node) {
                node.setUpdateType(newUpdateType);
            }
        });
        return this;
    }

    /**
     * If reducer returns an empty Maybe, nothing is added to the list.
     */
    public <T> List<T> reduce(final Function<EntityTreeNode, Maybe<T>> reducer) {
        final List<T> result = new ArrayList<>();
        reducer.apply(this)
                .ifPresent(new Consumer<T>() {
                    @Override
                    public void consume(final T value) {
                        result.add(value);
                    }
                });
        Sequence.of(children)
                .each(new BiConsumer<EntityTreeNode, Integer>() {
                    @Override
                    public void consume(EntityTreeNode child, Integer index) {
                        result.addAll(child.reduce(reducer));
                    }
                });
        return result;
    }

    public Maybe<EntityTreeNode> searchRecursively(EntityTo another, BiFunction<EntityTo, EntityTo, Boolean> matcher) {
        if (matcher.apply(entity, another)) {
            return Maybe.of(this);
        } else if (children.isEmpty()) {
            return Maybe.empty();
        } else {
            for (EntityTreeNode child : children) {
                Maybe<EntityTreeNode> match = child.searchRecursively(another, matcher);
                if (match.isPresent()) {
                    return match;
                }
            }
            return Maybe.empty();
        }
    }

    public void sortChildrenByName() {
        Collections.sort(children, new Comparator<EntityTreeNode>() {
            @Override
            public int compare(EntityTreeNode node1, EntityTreeNode node2) {
                return node1.entity.getName().compareTo(node2.entity.getName());
            }
        });
    }

    public void sortChildrenById() {
        Collections.sort(children, new Comparator<EntityTreeNode>() {
            @Override
            public int compare(EntityTreeNode node1, EntityTreeNode node2) {
                return node1.entity.getId().compareTo(node2.entity.getId());
            }
        });
    }
}
