package com.restlet.client.script;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.async.Promises;
import com.restlet.client.dao.ContextsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.log.LogService;
import com.restlet.client.model.EntityTo;
import com.restlet.client.model.context.ContextTo;
import com.restlet.client.net.http.request.HttpRequestTo;
import com.restlet.client.net.http.response.HttpResponseTo;
import com.restlet.client.net.response.ResponseBodyTo;
import com.restlet.client.platform.Base64;
import com.restlet.client.platform.CryptoApi;
import com.restlet.client.platform.RandomValueGenerator;
import com.restlet.client.platform.blob.BlobReader;
import com.restlet.client.platform.json.JsonEngine;
import com.restlet.client.script.method.Base64Method;
import com.restlet.client.script.method.ConcatMethod;
import com.restlet.client.script.method.HMACMethod;
import com.restlet.client.script.method.HashMethod;
import com.restlet.client.script.method.JsonPathMethod;
import com.restlet.client.script.method.LengthMethod;
import com.restlet.client.script.method.LowerMethod;
import com.restlet.client.script.method.MD5Method;
import com.restlet.client.script.method.RandomNumberMethod;
import com.restlet.client.script.method.StringMethod;
import com.restlet.client.script.method.SubStringMethod;
import com.restlet.client.script.method.TimestampMethod;
import com.restlet.client.script.method.UUIDGeneratorMethod;
import com.restlet.client.script.method.UpperMethod;
import com.restlet.client.script.parser.impl.ScriptParserImpl;
import com.restlet.client.script.parser.token.ScriptToken;
import com.restlet.client.script.runtime.ScriptMethod;
import com.restlet.client.script.runtime.ScriptRuntime;
import com.restlet.client.script.runtime.ScriptValue;
import com.restlet.client.script.runtime.impl.RepositoryRecordReference;
import com.restlet.client.script.runtime.impl.RequestContainerJsonReference;
import com.restlet.client.script.runtime.impl.RequestJSONReference;
import com.restlet.client.script.runtime.impl.ResponseJSONReference;
import com.restlet.client.script.value.ScriptJsonValue;
import com.restlet.client.script.value.ScriptNumberValue;
import com.restlet.client.script.value.ScriptStringValue;
import com.restlet.client.utils.BlobUtils;
import com.restlet.client.utils.ContextToHelper;
import com.restlet.client.utils.Holder;
import com.restlet.client.utils.MapBuilder;
import com.restlet.client.utils.Nullable;
import com.restlet.client.utils.Sequence;

import java.util.List;
import java.util.Map;

import static com.restlet.client.script.parser.token.ScriptTokenType.Expression;
import static com.restlet.client.script.parser.token.ScriptTokenType.LegacyContext;
import static com.restlet.client.script.parser.token.ScriptTokenType.Number;
import static com.restlet.client.script.parser.token.ScriptTokenType.Statement;
import static com.restlet.client.utils.Objects.isNullOrEmpty;
import static com.restlet.client.utils.StringUtils.format;
import static com.restlet.client.utils.StringUtils.isBlank;

public class ScriptEvaluator implements RepositoryReader, ScriptService, ScriptRuntime {

    private static final Predicate<ScriptToken> EXPRESSION_TOKEN_FILTER = new Predicate<ScriptToken>() {
        @Override
        public boolean test(ScriptToken token) {
            return token.getType() == Expression;
        }
    };
    private static final Predicate<ScriptToken> LEGACY_CONTENT_OR_NON_EMPTY_STATEMENT_FILTER = new Predicate<ScriptToken>() {
        @Override
        public boolean test(ScriptToken token) {
            return token.getType() == LegacyContext || token.getType() == Statement && !isNullOrEmpty(token.getChildren());
        }
    };
    private final JsonEngine jsonEngine;
    private final BlobReader blobReader;
    private final LogService logService;
    private final Function<String, Promise<HttpResponseTo>> lastResponseProvider;
    private final Function<ResponseBodyTo, Promise<ResponseBodyTo>> responseBodyProvider;
    private final Map<String, ScriptMethod> methods;

    public ScriptEvaluator(Base64 base64,
                           CryptoApi cryptoApi,
                           JsonEngine jsonEngine,
                           RandomValueGenerator randomValueGenerator,
                           BlobReader blobReader,
                           LogService logService,
                           Function<String, Promise<HttpResponseTo>> lastResponseProvider,
                           Function<ResponseBodyTo, Promise<ResponseBodyTo>> responseBodyProvider) {
        this.jsonEngine = jsonEngine;
        this.blobReader = blobReader;
        this.logService = logService;
        this.lastResponseProvider = lastResponseProvider;
        this.responseBodyProvider = responseBodyProvider;

        methods = MapBuilder.<String, ScriptMethod>builder()
                .put("random", new RandomNumberMethod(randomValueGenerator))
                .put("uuid", new UUIDGeneratorMethod(jsonEngine))
                .put("jsonPath", new JsonPathMethod(jsonEngine))
                .put("string", new StringMethod(jsonEngine))
                .put("substring", new SubStringMethod(jsonEngine, logService))
                .put("base64", new Base64Method(base64, jsonEngine))
                .put("md5", new MD5Method(cryptoApi, jsonEngine))
                .put("hmac", new HMACMethod(cryptoApi, jsonEngine, logService))
                .put("concat", new ConcatMethod(jsonEngine))
                .put("lower", new LowerMethod(jsonEngine))
                .put("upper", new UpperMethod(jsonEngine))
                .put("timestamp", new TimestampMethod())
                .put("length", new LengthMethod(jsonEngine))
                .put("size", new LengthMethod(jsonEngine))
                .put("sha", new HashMethod(cryptoApi, jsonEngine, logService))
                .build();
    }

    @Override
    public ScriptMethod getMethod(String name) {
        return methods.get(name.trim());
    }

    @Override
    public Promise<Void> restore() {
        return Promises.of();
    }

    @Override
    public Promise<Void> suspend() {
        return Promises.of();
    }

    @Override
    public Promise<String> evaluate(final String expression, ContextsDao contextsDao, RepositoryDao repositoryDao) {
        if (isBlank(expression)) {
            return Promises.of(expression);
        }

        final List<ScriptToken> tokens = new ScriptParserImpl().parse(expression);

        return evaluateTokens(tokens, contextsDao, repositoryDao)
                .map(new Function<List<ScriptValue>, String>() {
                    @Override
                    public String apply(List<ScriptValue> values) {
                        return buildExpressionResultFromTokensAndValues(expression, tokens, values);
                    }
                })
                .recoverOnReject(new PromiseHandler<Throwable>() {
                    @Override
                    public Object on(@Nullable Throwable error) {
                        logService.error(error.getMessage(), error);
                        return expression;
                    }
                });
    }

    @Override
    public Promise<ScriptValue> getVariable(final String path, final ContextsDao contextsDao, final RepositoryDao repositoryDao) {
        if (isBlank(path)) {
            return Promises.of();
        }

        return contextsDao.readSelected()
                .flatMap(new Function<ContextTo, Promise<ScriptValue>>() {
                    @Override
                    public Promise<ScriptValue> apply(ContextTo selectedContext) {
                        String computedValue;
                        if (selectedContext == null || (computedValue = ContextToHelper.expandVariables(selectedContext.getVariables(), path)) == null) {
                            return readFromRepo(path, null, repositoryDao, contextsDao);
                        }
                        final ScriptValue scriptStringValue = new ScriptStringValue(computedValue, jsonEngine);
                        return Promises.of(scriptStringValue);
                    }
                })
                .recoverOnReject(new PromiseHandler<Throwable>() {
                    @Override
                    public Object on(@Nullable final Throwable error) {
                        return null;
                    }
                });
    }

    private Promise<List<ScriptValue>> evaluateTokens(final List<ScriptToken> tokens, final ContextsDao contextsDao, final RepositoryDao repositoryDao) {
        if (isNullOrEmpty(tokens)) {
            return Promises.of();
        }

        List<Promise<ScriptValue>> tokensResolution = Sequence.of(tokens)
                .omitNulls()
                .filter(EXPRESSION_TOKEN_FILTER)
                .flatMap(new Function<ScriptToken, Iterable<Promise<ScriptValue>>>() {
                    @Override
                    public Iterable<Promise<ScriptValue>> apply(final ScriptToken token) {
                        return Sequence.of(token.getChildren())
                                .omitNulls()
                                .filter(LEGACY_CONTENT_OR_NON_EMPTY_STATEMENT_FILTER)
                                .map(new Function<ScriptToken, Promise<ScriptValue>>() {
                                    @Override
                                    public Promise<ScriptValue> apply(ScriptToken part) {
                                        if (part.getType() == LegacyContext) {
                                            return getVariable(part.getValue(), contextsDao, repositoryDao)
                                                    .doOnResolve(tokenValueSetter(token));
                                        } else {
                                            return evaluateStatement(part, contextsDao, repositoryDao)
                                                    .doOnResolve(tokenValueSetter(token));
                                        }
                                    }
                                })
                                .toList();
                    }
                })
                .toList();

        return Promises.all(tokensResolution);
    }

    private Consumer<ScriptValue> tokenValueSetter(final ScriptToken token) {
        return new Consumer<ScriptValue>() {
            @Override
            public void consume(ScriptValue value) {
                token.setValue(value != null ? value.toString() : null);
            }
        };
    }

    private Promise<ScriptValue> evaluateStatement(ScriptToken statement, final ContextsDao contextsDao, final RepositoryDao repositoryDao) {

        if (statement == null || isNullOrEmpty(statement.getChildren())) {
            return Promises.of();
        }

        List<ScriptToken> statementParts = Sequence.of(statement.getChildren()).omitNulls().toList();

        Promise<ScriptValue> promise = doStatementItem(statementParts.iterator().next(), contextsDao, repositoryDao);

        if (promise == null) {
            return Promises.of();
        }

        for (int i = 1; i < statementParts.size(); i++) {

            final ScriptToken nextItem = statement.getChildren().get(i);

            switch (nextItem.getType()) {
                case String:
                case Reference:
                    promise = promise.flatMap(new Function<ScriptValue, Promise<ScriptValue>>() {
                        @Override
                        public Promise<ScriptValue> apply(ScriptValue scriptValue) {
                            if (scriptValue == null) {
                                return Promises.of();
                            }

                            return scriptValue.select(nextItem.getValue());
                        }
                    });
                    break;

                case IndexReference:
                    promise = promise
                            .flatMap(new Function<ScriptValue, Promise<ScriptValue>>() {
                                @Override
                                public Promise<ScriptValue> apply(ScriptValue previousValue) {
                                    return previousValue == null ?
                                            Promises.<ScriptValue>of() :
                                            previousValue.select(nextItem.getValue());
                                }
                            })
                            .flatMap(resolveChildrenScriptValues(nextItem, contextsDao, repositoryDao));
                    break;

                case Method:

                    final ScriptMethod method = getMethod(nextItem.getValue());
                    if (method == null) {
                        return Promises.of();
                    }

                    promise = promise.flatMap(new Function<ScriptValue, Promise<ScriptValue>>() {
                        @Override
                        public Promise<ScriptValue> apply(final ScriptValue prevValue) {
                            if (prevValue == null) {
                                return Promises.of();
                            }

                            // do method arguments
                            return executeMethod(method, prevValue, nextItem, contextsDao, repositoryDao);
                        }
                    });

                    break;

                default:
                    return Promises.of();
            }

        }

        if (promise == null) {
            return Promises.of();
        }

        return promise;
    }

    private Promise<ScriptValue> doStatementItem(final ScriptToken scriptToken, final ContextsDao contextsDao, final RepositoryDao repositoryDao) {

        if (scriptToken == null) {
            return Promises.of();
        }

        if (scriptToken.getType() == Number) {
            try {
                final ScriptValue value = new ScriptNumberValue(Long.valueOf(scriptToken.getValue()));
                return Promises.of(value);
            } catch (NumberFormatException ignored) {
                logService.error(format("Provided number %s is not a valid long", scriptToken.getValue()), ignored);
            }
        }

        switch (scriptToken.getType()) {
            case Number:
            case String:
                return getVariable(scriptToken.getValue(), contextsDao, repositoryDao)
                        .map(new Function<ScriptValue, ScriptValue>() {
                            @Override
                            public ScriptValue apply(ScriptValue originalValue) {
                                return originalValue == null ?
                                        new ScriptStringValue(scriptToken.getValue(), jsonEngine) :
                                        originalValue;
                            }
                        });

            case Reference:
                return getVariable(scriptToken.getValue(), contextsDao, repositoryDao);

            case IndexReference:
                return getVariable(scriptToken.getValue(), contextsDao, repositoryDao)
                        .flatMap(resolveChildrenScriptValues(scriptToken, contextsDao, repositoryDao));

            case Method:
                final ScriptMethod method = getMethod(scriptToken.getValue());
                if (method != null) {
                    return executeMethod(method, null, scriptToken, contextsDao, repositoryDao);
                }
                break;

            default:
                break;
        }
        return null;
    }

    private Function<ScriptValue, Promise<ScriptValue>> resolveChildrenScriptValues(final ScriptToken scriptToken, final ContextsDao contextsDao, final RepositoryDao repositoryDao) {
        return new Function<ScriptValue, Promise<ScriptValue>>() {
            @Override
            public Promise<ScriptValue> apply(final ScriptValue value) {
                if (value == null) {
                    return null;
                }

                Promise<ScriptValue> promise = Promises.of(value);

                if (!isNullOrEmpty(scriptToken.getChildren())) {
                    for (final ScriptToken token : scriptToken.getChildren()) {
                        final Holder<ScriptValue> previousChildrenTokenValue = Holder.empty();
                        promise = promise.assignTo(previousChildrenTokenValue)
                                .flatMap(
                                        evaluateStatement(token, contextsDao, repositoryDao)
                                                .doOnResolve(tokenValueSetter(token))
                                )
                                .flatMap(new Function<ScriptValue, Promise<ScriptValue>>() {
                                    @Override
                                    public Promise<ScriptValue> apply(ScriptValue input) {
                                        if (input == null) {
                                            return Promises.of(previousChildrenTokenValue.value);
                                        }

                                        return previousChildrenTokenValue.value.select(input.toString());
                                    }
                                });
                    }
                }
                return promise;
            }
        };
    }

    @Override
    public Promise<ScriptValue> readFromRepo(final String path, final EntityTo parent, final RepositoryDao repositoryDao, final ContextsDao contextsDao) {

        if (isBlank(path)) {
            return Promises.of();
        }

        return repositoryDao.find(path, parent)
                .flatMap(new Function<EntityTo, Promise<ScriptValue>>() {
                    @Override
                    public Promise<ScriptValue> apply(EntityTo entity) {
                        if (entity == null) {
                            return Promises.of();
                        }

                        switch (entity.getType()) {

                            case Project:
                            case Service:
                            case Scenario:
                                final ScriptValue reference = new RepositoryRecordReference(ScriptEvaluator.this, entity, contextsDao, repositoryDao);
                                return Promises.of(reference);

                            case Request:

                                final RequestContainerJsonReference requestContainerJsonReference = new RequestContainerJsonReference(jsonEngine.newJsonObject(), jsonEngine);
                                requestContainerJsonReference.setName(entity.getName());

                                Promise<HttpResponseTo> lastResponse = lastResponseProvider.apply(entity.getId());

                                if (lastResponse == null) {
                                    final ScriptValue scriptJsonValue = new ScriptJsonValue(requestContainerJsonReference.asJSON(), jsonEngine);
                                    return Promises.of(scriptJsonValue);
                                }

                                final Holder<ResponseJSONReference> responseReferenceHolder = Holder.empty();
                                HttpRequestTo requestTo = (HttpRequestTo) entity;

                                return RequestJSONReference.create(jsonEngine, ScriptEvaluator.this, contextsDao, repositoryDao,
                                        requestTo.getMethod(),
                                        requestTo.getHeaders(),
                                        requestTo.getUri(),
                                        requestTo.getBody())
                                        .doOnResolve(new Consumer<RequestJSONReference>() {
                                            @Override
                                            public void consume(RequestJSONReference requestJsonReference) {
                                                requestContainerJsonReference.setRequest(requestJsonReference);
                                            }
                                        })
                                        .flatMap(lastResponse)
                                        .flatMap(new Function<HttpResponseTo, Promise<ResponseBodyTo>>() {

                                            @Override
                                            public Promise<ResponseBodyTo> apply(HttpResponseTo result) {
                                                responseReferenceHolder.value = new ResponseJSONReference(jsonEngine.newJsonObject(), jsonEngine);

                                                if (result == null) {
                                                    return Promises.of();
                                                }

                                                responseReferenceHolder.value.setHeaders(result.getHeaders());
                                                if (result.getStatus() != null) {
                                                    responseReferenceHolder.value.setStatus(result.getStatus().getCode(), result.getStatus().getMessage());
                                                }

                                                if (result.getBody() == null) {
                                                    return Promises.of();
                                                } else if (result.getBody().getBlob() == null) {
                                                    return responseBodyProvider.apply(result.getBody());
                                                } else {
                                                    return Promises.of(result.getBody());
                                                }

                                            }
                                        })
                                        .flatMap(new Function<ResponseBodyTo, Promise<String>>() {
                                            @Override
                                            public Promise<String> apply(ResponseBodyTo responseBodyTo) {
                                                if (responseBodyTo == null) {
                                                    return Promises.of();
                                                }
                                                return BlobUtils.read(blobReader, responseBodyTo.getBlob());
                                            }
                                        })
                                        .map(new Function<String, ScriptValue>() {
                                            @Override
                                            public ScriptValue apply(String readBlob) {
                                                responseReferenceHolder.value.setBody(readBlob);
                                                return new ScriptJsonValue(requestContainerJsonReference.asJSON(), jsonEngine);
                                            }
                                        })
                                        .recoverOnReject(new PromiseHandler<Throwable>() {
                                            @Override
                                            public Object on(@Nullable Throwable error) {
                                                return new ScriptJsonValue(requestContainerJsonReference.asJSON(), jsonEngine);
                                            }
                                        })
                                        .doFinally(new Consumer<Deferred>() {
                                            @Override
                                            public void consume(Deferred deferred) {
                                                requestContainerJsonReference.setResponse(responseReferenceHolder.value);
                                            }
                                        });

                            default:
                                break;
                        }

                        return Promises.of();
                    }
                });
    }

    private Promise<ScriptValue> executeMethod(final ScriptMethod method, final ScriptValue value, ScriptToken arguments, final ContextsDao contextsDao, final RepositoryDao repositoryDao) {
        return Promises.all(
                Sequence.of(arguments.getChildren())
                        .map(new Function<ScriptToken, Promise<ScriptValue>>() {
                            @Override
                            public Promise<ScriptValue> apply(ScriptToken argToken) {
                                return evaluateStatement(argToken, contextsDao, repositoryDao)
                                        .doOnResolve(tokenValueSetter(argToken));
                            }
                        }).toList())
                .flatMap(new Function<List<ScriptValue>, Promise<ScriptValue>>() {
                    @Override
                    public Promise<ScriptValue> apply(List<ScriptValue> arguments) {
                        return method.execute(value, arguments);
                    }
                });
    }

    private String buildExpressionResultFromTokensAndValues(String expression, List<ScriptToken> tokens, List<ScriptValue> values) {
        if (isNullOrEmpty(values)) {
            return expression;
        }

        StringBuilder output = new StringBuilder();

        for (ScriptToken token : tokens) {
            switch (token.getType()) {
                case Expression:
                    String value = token.getValue();
                    if (value != null) {
                        output.append(value);
                    } else {
                        output.append(token.getSource());
                    }
                    break;

                default:
                    output.append(token.getSource());
                    break;
            }
        }
        return output.toString();
    }
}
