package com.restlet.client.script.runtime.impl;

import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.dao.ContextsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.Consumer;
import com.restlet.client.net.http.request.HttpMethodDefinition;
import com.restlet.client.net.request.RequestBodyTo;
import com.restlet.client.net.request.RequestHeaderTo;
import com.restlet.client.net.uri.UriTo;
import com.restlet.client.net.uri.UriUtils;
import com.restlet.client.platform.json.JsonArray;
import com.restlet.client.platform.json.JsonEngine;
import com.restlet.client.platform.json.JsonException;
import com.restlet.client.platform.json.JsonObject;
import com.restlet.client.platform.json.JsonValue;
import com.restlet.client.script.ScriptEvaluator;
import com.restlet.client.function.Function;
import com.restlet.client.utils.StringUtils;

import java.util.List;

import static com.restlet.client.net.uri.UriUtils.SchemePrintOption.PRINT_SCHEME;

public class RequestJSONReference {

    private JsonObject root;
    private JsonEngine jsonEngine;
    private ScriptEvaluator scriptEvaluator;
    private ContextsDao contextsDao;
    private RepositoryDao repositoryDao;

    private RequestJSONReference(JsonEngine jsonEngine,
                                 ScriptEvaluator scriptEvaluator,
                                 ContextsDao contextsDao,
                                 RepositoryDao repositoryDao) {
        this.root = jsonEngine.newJsonObject();
        this.jsonEngine = jsonEngine;
        this.scriptEvaluator = scriptEvaluator;
        this.contextsDao = contextsDao;
        this.repositoryDao = repositoryDao;
    }

    public void setMethod(final HttpMethodDefinition method) {
        if (method != null) {
            root.put("method", jsonEngine.newJsonString(method.getName()));
        } else {
            root.put("method", null);
        }
    }

    public void setHeaders(final List<RequestHeaderTo> headers) {
        if (headers != null) {
            JsonObject refHeaders = jsonEngine.newJsonObject();
            for (RequestHeaderTo reqHeader : headers) {
                if (reqHeader == null || !reqHeader.isEnabled() || StringUtils.isBlank(reqHeader.getName()) || reqHeader.getValue() == null) {
                    continue;
                }

                JsonValue refHeader = refHeaders.get(reqHeader.getName().toLowerCase());
                if (refHeader == null) {
                    refHeaders.put(reqHeader.getName().toLowerCase(), jsonEngine.newJsonString(reqHeader.getValue()));

                } else if (refHeader.isArray() != null) {
                    refHeader.isArray().set(refHeader.isArray().size(), jsonEngine.newJsonString(reqHeader.getValue()));

                } else if (refHeader.isString() != null) {
                    JsonArray refHeadersArray = jsonEngine.newJsonArray();
                    refHeadersArray.set(0, refHeader);
                    refHeadersArray.set(1, jsonEngine.newJsonString(reqHeader.getValue()));
                    refHeaders.put(reqHeader.getName().toLowerCase(), refHeadersArray);
                }
            }
            root.put("headers", refHeaders);
        } else {
            root.put("headers", null);
        }
    }

    public JsonValue asJSON() {
        return root;
    }

    public void setUri(UriTo uri) {
        if (uri == null) {
            root.put("uri", null);
            return;
        }
        root.put("uri", jsonEngine.newJsonString(UriUtils.printUri(uri, PRINT_SCHEME)));
    }

    public Promise<Void> setBody(final RequestBodyTo body) {
        if (body != null && body.getBodyType() != null) {
            switch (body.getBodyType()) {
                case File:
                case Form:
                    break;

                case Text:
                    if (!StringUtils.isBlank(body.getTextBody())) {
                        return scriptEvaluator.evaluate(body.getTextBody(), contextsDao, repositoryDao)
                                .doOnResolve(new Consumer<String>() {
                                    @Override
                                    public void consume(String parsedValue) {
                                        JsonValue json = null;
                                        try {
                                            json = RequestJSONReference.this.jsonEngine.fromJson(parsedValue);
                                        } catch (JsonException ignored) {
                                        }

                                        if (json != null) {
                                            RequestJSONReference.this.root.put("body", json);
                                            return;
                                        }
                                        RequestJSONReference.this.root.put("body", RequestJSONReference.this.jsonEngine.newJsonString(body.getTextBody()));
                                    }
                                })
                                .castToVoidPromise();

                    }
                    break;
            }
        }
        return Promises.of();
    }

    public static Promise<RequestJSONReference> create(JsonEngine jsonEngine,
                                                       ScriptEvaluator scriptEvaluator,
                                                       ContextsDao contextsDao,
                                                       RepositoryDao repositoryDao,
                                                       HttpMethodDefinition method,
                                                       List<RequestHeaderTo> headers,
                                                       UriTo uri,
                                                       RequestBodyTo body) {
        final RequestJSONReference requestJSONReference = new RequestJSONReference(jsonEngine, scriptEvaluator, contextsDao, repositoryDao);
        requestJSONReference.setMethod(method);
        requestJSONReference.setHeaders(headers);
        requestJSONReference.setUri(uri);
        return requestJSONReference.setBody(body)
                .map(new Function<Void, RequestJSONReference>() {
                    @Override
                    public RequestJSONReference apply(Void whatever) {
                        return requestJSONReference;
                    }
                });

    }
}
