package com.restlet.client.tests.asserts.impl;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.html.HtmlKit;
import com.restlet.client.net.http.HttpDate;
import com.restlet.client.net.http.response.HttpResponseTo;
import com.restlet.client.platform.blob.BlobReader;
import com.restlet.client.platform.json.JsonArray;
import com.restlet.client.platform.json.JsonEngine;
import com.restlet.client.platform.json.JsonException;
import com.restlet.client.platform.json.JsonObject;
import com.restlet.client.platform.json.JsonValue;
import com.restlet.client.tests.asserts.AssertionResultTo;
import com.restlet.client.tests.asserts.AssertionResultTo.State;
import com.restlet.client.tests.asserts.ResponseBodyAssertionEngine;
import com.restlet.client.utils.StringUtils;

import java.util.Date;
import java.util.Objects;

import static com.restlet.client.utils.StringUtils.firstNotBlank;
import static com.restlet.client.utils.StringUtils.isBlank;

public class JsonAssertionEngineImpl implements ResponseBodyAssertionEngine {

    private final JsonEngine jsonEngine;
    private final HttpDate httpDate;
    private final BlobReader blobReader;
    private final HtmlKit htmlKit;

    public JsonAssertionEngineImpl(JsonEngine jsonEngine, HttpDate httpDate, BlobReader blobReader, HtmlKit htmlKit) {
        super();
        this.jsonEngine = jsonEngine;
        this.httpDate = httpDate;
        this.blobReader = blobReader;
        this.htmlKit = htmlKit;
    }

    private boolean matchJson(String json, AssertionResultTo assertion) {

        String stringPath = firstNotBlank(assertion.getEvaluatedPath(), assertion.getPath()).orElse(null);

        if (isBlank(stringPath)) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> is required.");
            return false;
        }

        if (!stringPath.startsWith("$")) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("&lt;" + stringPath + "&gt; is not valid <a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a>.");
            return false;
        }

        try {
            JsonValue root = jsonEngine.fromJson(json);

            if (root == null) {
                assertion.setState(State.Failure);
                assertion.addFailureMessage("Response body is not valid Json.");
                return false;
            }

            if (root.isObject() == null && root.isArray() == null) {
                assertion.setState(State.Failure);
                assertion.addFailureMessage("Response body is not Json Object.");
                return false;
            }

            JsonValue sourceValue = root.isObject() != null
                    ? extract(root.isObject(), stringPath, assertion)
                    : extract(root.isArray(), stringPath, assertion);

            JsonValue expectedValue;

            String stringValue = firstNotBlank(assertion.getEvaluatedValue(), assertion.getValue()).orElse(null);

            if (!isBlank(stringValue)) {
                try {

                    expectedValue = jsonEngine.fromJson(stringValue);

                } catch (JsonException e) {

                    expectedValue = jsonEngine.newJsonString(stringValue);

                }
            } else {
                expectedValue = jsonEngine.newJsonString("");
            }

            if (sourceValue != null && sourceValue.isNull() != null) {
                sourceValue = null;
            }
            if (expectedValue != null && expectedValue.isNull() != null) {
                expectedValue = null;
            }

            Boolean validated = validate(assertion, sourceValue, expectedValue);

            return (validated != null) ? validated : false;
        } catch (Throwable e) {
            assertion.setState(State.Failure);
            assertion.addFailureMessage("Response body is not valid JSON. [" + htmlKit.escape(e.getMessage()) + "]");
        }
        return false;
    }

    public Boolean validate(AssertionResultTo assertion, JsonValue sourceValue, JsonValue expectedValue) {
        switch (assertion.getComparison()) {
            case Equals:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (equals(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Expected sourceValue: &lt;" + htmlEscape(expectedValue, "null") + "&gt; but was: &lt;" + htmlEscape(sourceValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case DoesNotEqual:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (!equals(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value: &lt;" + htmlEscape(sourceValue, "null") + "&gt is the same;.");
                }
                return State.Ok == assertion.getState();

            case Exists:
                if (sourceValue != null) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");
                }
                return State.Ok == assertion.getState();

            case DoesNotExist:
                if (sourceValue == null) {
                    assertion.setState(State.Ok);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target exists, &lt;" + htmlEscape(sourceValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case Contains:
                if (sourceValue != null && expectedValue != null && contains(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; does not contain &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case DoesNotContain:
                if (sourceValue != null && expectedValue != null && !contains(sourceValue, expectedValue)) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");

                } else if (expectedValue != null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; contains &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case Less:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) < 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not less than &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case LessOrEqual:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) <= 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not less or equal to &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case Greater:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) > 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not greater than &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            case GreateOrEqual:
                warnUserTypesDiffer(assertion, sourceValue, expectedValue);
                if (sourceValue != null && expectedValue != null && compare(sourceValue, expectedValue) >= 0) {
                    assertion.setState(State.Ok);

                } else if (sourceValue == null) {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> target does not exist or is a null.");

                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Value &lt;" + htmlEscape(sourceValue) + "&gt; is not greater or equal to &lt;" + htmlEscape(expectedValue, "null") + "&gt;.");
                }
                return State.Ok == assertion.getState();

            default:
                break;
        }
        return null;
    }

    private String htmlEscape(JsonValue value, String defaultValue) {
        return value == null ? defaultValue : htmlKit.escape(value.toString());
    }

    private String htmlEscape(JsonValue value) {
        return htmlEscape(value, null);
    }

    private JsonValue extract(JsonArray root, String selector, AssertionResultTo assertion) {
        try {
            JsonValue val = jsonEngine.jsonPath(root, selector);

            if (val == null) {
                return root;
            }
            return val;

        } catch (JsonException e) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> cannot be evaluated.");
        }
        return null;
    }

    private JsonValue extract(JsonObject root, String selector, AssertionResultTo assertion) {
        try {
            return jsonEngine.jsonPath(root, selector);

        } catch (JsonException e) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> cannot be evaluated.");
        }
        return null;
    }

    private boolean haveSameType(JsonValue v1, JsonValue v2) {
        return areNullValues(v1, v2)
                || areBooleanValues(v1, v2)
                || areNumberValues(v1, v2)
                || areArrayValues(v1, v2)
                || areObjectValues(v1, v2)
                || areStringValues(v1, v2);
    }

    private String getType(JsonValue value) {

        if (value.isNull() != null) {
            return "null";
        } else if (value.isArray() != null) {
            return "array";
        } else if (value.isBoolean() != null) {
            return "boolean";
        } else if (value.isNumber() != null) {
            return "number";
        } else if (value.isString() != null) {
            return "string";
        } else if (value.isObject() != null) {
            return "object";
        }
        return null;
    }

    private void warnUserTypesDiffer(AssertionResultTo assertion, JsonValue value, JsonValue expected) {
        if (expected == null || value == null) {
            return;
        }
        if (!haveSameType(value, expected)) {
            assertion.addWarningMessage("Assertion compares values as strings (<a href=\"https://en.wikipedia.org/wiki/Lexicographical_order\" target=\"_blank\">lexicographically</a>) since &lt;" + htmlEscape(value) + "&gt; and &lt;" + htmlEscape(expected, "null") + "&gt; have different types (&lt;" + htmlEscape(value) + "&gt; has type \"" + getType(value) + "\" and &lt;" + htmlEscape(expected, "null") + "&gt; has type \"" + getType(expected) + "\")");
        }
    }

    public int compare(JsonValue v1, JsonValue v2) {

        if (v1 == null && v2 == null) {
            return 0;
        } else if (v1 == null) {
            return -1;
        } else if (v2 == null) {
            return 1;
        }

        if (areNullValues(v1, v2)) {
            return 0;
        }

        if (areBooleanValues(v1, v2)) {
            if (v1.isBoolean().booleanValue() == v2.isBoolean().booleanValue()) {
                return 0;
            }
            if (v1.isBoolean().booleanValue()) {
                return 1;
            }
            return -1;
        }

        if (areNumberValues(v1, v2)) {
            if (v1.isNumber().doubleValue() == v2.isNumber().doubleValue()) {
                return 0;
            }
            if (v1.isNumber().doubleValue() > v2.isNumber().doubleValue()) {
                return 1;
            }
            return -1;
        }

        if (areArrayValues(v1, v2)) {
            if (v1.isArray().size() < v2.isArray().size()) {
                return -1;
            }
            if (v1.isArray().size() > v2.isArray().size()) {
                return 1;
            }

            boolean equals = true;

            for (int i = 0; i < v1.isArray().size(); i++) {
                equals = equals(v1.isArray().get(i), v2.isArray().get(i));
                if (!equals) {
                    break;
                }
            }

            if (equals) {
                return 0;
            }

        }

        if (areObjectValues(v1, v2)) {
            if (v1.isObject().keySet().size() < v2.isObject().keySet().size()) {
                return -1;
            }
            if (v1.isObject().keySet().size() > v2.isObject().keySet().size()) {
                return 1;
            }

            boolean equals = true;
            for (String key : v1.isObject().keySet()) {
                equals = v2.isObject().containsKey(key) && equals(v1.isObject().get(key), v2.isObject().get(key));
                if (!equals) {
                    break;
                }
            }

            if (equals) {
                return 0;
            }
        }

        if (areStringValues(v1, v2)) {
            String s1 = v1.isString().stringValue();
            String s2 = v2.isString().stringValue();

            try {
                Long l1 = Long.valueOf(s1);
                Long l2 = Long.valueOf(s2);

                if (l1 != null && l2 != null) {
                    return l1.compareTo(l2);
                }

            } catch (Exception e) {
                // do nothing, let's try to convert to another type
            }

            try {
                Double d1 = Double.valueOf(s1);
                Double d2 = Double.valueOf(s2);
                if (d1 != null && d2 != null) {
                    return d1.compareTo(d2);
                }

            } catch (Exception e) {
                // do nothing, let's try to convert to another type
            }

            Date d1 = httpDate.valueOf(s1);
            Date d2 = httpDate.valueOf(s2);

            if (d1 != null && d2 != null) {
                return d1.compareTo(d2);
            }

            return s1.compareToIgnoreCase(s2);
        }

        String s1 = v1.isString() != null ? v1.isString().stringValue() : v1.toString();
        String s2 = v2.isString() != null ? v2.isString().stringValue() : v2.toString();

        return (s1 != null && s2 != null) ? s1.compareToIgnoreCase(s2) : (s1 == null ? -1 : 1);
    }

    private boolean areStringValues(JsonValue v1, JsonValue v2) {
        return v1.isString() != null && v2.isString() != null;
    }

    private boolean areObjectValues(JsonValue v1, JsonValue v2) {
        return v1.isObject() != null && v2.isObject() != null;
    }

    private boolean areArrayValues(JsonValue v1, JsonValue v2) {
        return v1.isArray() != null && v2.isArray() != null;
    }

    private boolean areNumberValues(JsonValue v1, JsonValue v2) {
        return v1.isNumber() != null && v2.isNumber() != null;
    }

    private boolean areBooleanValues(JsonValue v1, JsonValue v2) {
        return v1.isBoolean() != null && v2.isBoolean() != null;
    }

    private boolean areNullValues(JsonValue v1, JsonValue v2) {
        return v1.isNull() != null && v2.isNull() != null;
    }

    private boolean equals(JsonValue sourceValue, JsonValue expectedValue) {
        if (sourceValue != null && expectedValue != null) {

            sourceValue = unwrapFirstArrayElementIfTypeMismatch(sourceValue, expectedValue);

            if (areNullValues(sourceValue, expectedValue)) {
                return true;
            }

            if (areBooleanValues(sourceValue, expectedValue)) {
                return sourceValue.isBoolean().booleanValue() == expectedValue.isBoolean().booleanValue();
            }

            if (areNumberValues(sourceValue, expectedValue)) {
                return sourceValue.isNumber().doubleValue() == expectedValue.isNumber().doubleValue();
            }

            if (areArrayValues(sourceValue, expectedValue)) {
                if (Objects.equals(sourceValue.isArray().size(), expectedValue.isArray().size())) {
                    for (int i = 0; i < sourceValue.isArray().size(); i++) {
                        if (!equals(sourceValue.isArray().get(i), expectedValue.isArray().get(i))) {
                            return false;
                        }
                    }
                    return true;
                }
                return false;
            }

            if (areObjectValues(sourceValue, expectedValue)) {
                if (sourceValue.isObject().keySet().size() == expectedValue.isObject().keySet().size()) {
                    for (String key : sourceValue.isObject().keySet()) {
                        if (!expectedValue.isObject().containsKey(key)
                                || !equals(sourceValue.isObject().get(key), expectedValue.isObject().get(key))) {
                            return false;
                        }
                    }
                    return true;
                }
                return false;
            }

            if (areStringValues(sourceValue, expectedValue)) {
                String s1 = sourceValue.isString().stringValue();
                String s2 = expectedValue.isString().stringValue();

                Date d1 = httpDate.valueOf(s1);
                Date d2 = httpDate.valueOf(s2);

                if (d1 != null && d2 != null) {
                    return d1.equals(d2);
                }
                return s1.equalsIgnoreCase(s2);
            }

            String s1 = sourceValue.isString() != null ? sourceValue.isString().stringValue() : sourceValue.toString();
            String s2 = expectedValue.isString() != null ? expectedValue.isString().stringValue() : expectedValue.toString();
            return (s1 != null && s2 != null) && s1.equalsIgnoreCase(s2);
        }
        return (sourceValue == null && expectedValue == null);
    }

    private JsonValue unwrapFirstArrayElementIfTypeMismatch(JsonValue sourceValue, JsonValue expectedValue) {
        if (sourceValue.isArray() != null && sourceValue.isArray().size() == 1 && expectedValue.isArray() == null) {
            sourceValue = sourceValue.isArray().get(0);
        }
        return sourceValue;
    }

    private boolean contains(JsonValue v1, JsonValue v2) {
        if (v1 != null && v2 != null) {
            String s1 = v1.isString() != null ? v1.isString().stringValue() : v1.toString();
            String s2 = v2.isString() != null ? v2.isString().stringValue() : v2.toString();

            if (s1 != null && s2 != null) {
                return s1.toLowerCase().contains(s2.toLowerCase());
            }
        }
        return false;
    }

    @Override
    public Promise<Void> checkBody(final AssertionResultTo assertion, final HttpResponseTo response) {
        if (response == null) {
            return Promises.of();
        }

        if (isBlank(assertion.getPath())) {
            assertion.setState(State.Error);
            assertion.addErrorMessage("<a href=\"http://goessner.net/articles/JsonPath/\" target=\"_blank\">JSON Path</a> is required.");
            return Promises.of();
        }

        if (response.getBody() == null || response.getBody().getBlob() == null || response.getBody().getBlob().size() <= 0) {
            assertion.setState(State.Failure);
            assertion.addFailureMessage("Response body does not exist.");
            return Promises.of();
        }

        final Deferred<Void> deferred = new DeferredImpl<>();

        blobReader.readAsText(response.getBody().getBlob(), new BlobReader.Handler<String>() {

            @Override
            public void onLoaded(String responseBodyAsText) {
                if (!StringUtils.isBlank(responseBodyAsText)) {
                    assertion.setActualValue(responseBodyAsText);
                    matchJson(responseBodyAsText, assertion);
                } else {
                    assertion.setState(State.Failure);
                    assertion.addFailureMessage("Response body is not JSON.");
                }
                deferred.resolve();
            }

            @Override
            public void onError(Object o) {
                deferred.resolve();
            }
        });
        return deferred.promise();
    }
}
