package com.restlet.client.tests.impl;

import com.restlet.client.async.AsyncStreamWithResult;
import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.async.Promises;
import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.async.impl.WritableStreamWithResult;
import com.restlet.client.dao.ContextsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.net.http.HttpClient;
import com.restlet.client.net.http.HttpClientResultTo;
import com.restlet.client.net.http.HttpStatusCodeRange;
import com.restlet.client.net.http.request.HttpRequestTo;
import com.restlet.client.net.http.request.impl.EffectiveHttpRequestTo;
import com.restlet.client.script.ScriptService;
import com.restlet.client.script.impl.BaseRequestEvaluator;
import com.restlet.client.tests.ScenarioTo;
import com.restlet.client.tests.TestCaseResult;
import com.restlet.client.tests.TestResult;
import com.restlet.client.tests.TestResult.State;
import com.restlet.client.tests.TestRunner;
import com.restlet.client.tests.asserts.AssertionEngine;
import com.restlet.client.tests.asserts.AssertionResultTo;
import com.restlet.client.tests.asserts.AssertionTo;
import com.restlet.client.utils.EntityUtils;
import com.restlet.client.utils.Sequence;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import static com.restlet.client.net.http.HttpStatusCodeRange.ZERO;
import static com.restlet.client.utils.Objects.isNullOrEmpty;

public class TestRunnerImpl implements TestRunner {

    private static final List<AssertionResultTo.State> assertionStatesThatAlterRequestState = Arrays.asList(AssertionResultTo.State.Error, AssertionResultTo.State.Failure, AssertionResultTo.State.Ok);

    private final HttpClient httpClient;
    private final BaseRequestEvaluator requestEvaluator;
    private final AssertionEngine assertionEngine;
    private final HistoryCallback callback;

    private ContextsDao contextsDao;
    private RepositoryDao repositoryDao;
    private TestCaseResultStream stream;
    private List<HttpStatusCodeRange> acceptedCodeRanges = HttpStatusCodeRange.REQUESTS_VIEW_CODE_RANGES;

    protected TestRunnerImpl(HttpClient httpClient, AssertionEngine assertionEngine, final ScriptService script,
                             HistoryCallback callback, final ContextsDao contextsDao, final RepositoryDao repositoryDao,
                             final Function<String, String> queryParameterDecoder,
                             final Function<String, String> queryParameterEncoder, List<HttpStatusCodeRange> acceptedCodeRanges) {
        this(httpClient, assertionEngine, script, callback, contextsDao, repositoryDao, queryParameterDecoder, queryParameterEncoder);
        this.acceptedCodeRanges = acceptedCodeRanges;
    }

    protected TestRunnerImpl(HttpClient httpClient, AssertionEngine assertionEngine, final ScriptService scriptService,
                             HistoryCallback callback, final ContextsDao contextsDao, final RepositoryDao repositoryDao,
                             final Function<String, String> queryParameterDecoder,
                             final Function<String, String> queryParameterEncoder) {
        this.httpClient = httpClient;
        this.assertionEngine = assertionEngine;
        this.callback = callback;
        this.contextsDao = contextsDao;
        this.repositoryDao = repositoryDao;
        this.requestEvaluator = new BaseRequestEvaluator(contextsDao, scriptService) {

            @Override
            protected String encodeQueryString(String query) {
                return queryParameterEncoder.apply(query);
            }

            @Override
            protected String decodeQueryString(String query) {
                return queryParameterDecoder.apply(query);
            }
        };
    }

    @Override
    public AsyncStreamWithResult<TestResult, TestCaseResult> run(final ScenarioTo testScenario, final Consumer<HttpClientResultTo> onRequestComplete) {
        if (testScenario == null) {
            throw new IllegalArgumentException("The 'testCase' argument cannot be a null.");
        }

        // Initialize scenario
        final TestCaseResult scenarioResult = new TestCaseResultImpl()
                .setName(testScenario.getName())
                .setFullName(EntityUtils.generateDottedName(testScenario))
                .setState(State.InProgress);

        stream = new TestCaseResultStream(scenarioResult);

        if (isNullOrEmpty(testScenario.getHttpRequests())) {
            scenarioResult.setState(State.Ok);
            return stream.closeIfStillOpened(scenarioResult);
        }

        for (HttpRequestTo request : testScenario.getHttpRequests()) {

            TestResultImpl testResult = new TestResultImpl();
            testResult.setState(State.InProgress);
            testResult.setHttpRequest(request);

            scenarioResult.getTestResults().add(testResult);
        }

        Promise<Void> promise = Promises.of();

        // run each test
        for (final TestResult testResult : scenarioResult.getTestResults()) {
            promise = run(promise, testResult, onRequestComplete).doOnResolve(new Consumer<Void>() {
                @Override
                public void consume(Void whatever) {
                    if (stream.isClosed()) {
                        return;
                    }

                    stream.accept(testResult);
                    if (State.Failure == testResult.getState() || State.Error == testResult.getState()) {
                        stream.closeIfStillOpened(scenarioResult);
                    }
                }
            });
        }

        // compute the state of the whole scenario
        promise.doOnResolve(new Consumer<Void>() {
            @Override
            public void consume(Void whatever) {
                stream.closeIfStillOpened(scenarioResult);
            }
        });

        return stream;
    }

    private Promise<Void> run(Promise<Void> promise, final TestResult result, final Consumer<HttpClientResultTo> onRequestComplete) {

        if (stream.isClosed()) {
            return Promises.of();
        }

        final Deferred<Void> deferred = new DeferredImpl<>();

        promise.then(new PromiseHandler<Void>() {

            @Override
            public Object on(Void whatever) {

                if (stream.isClosed()) {
                    deferred.resolve();
                    return Promises.of();
                }

                return requestEvaluator.effective(result.getHttpRequest(), repositoryDao).then(new PromiseHandler<EffectiveHttpRequestTo>() {

                    @Override
                    public Object on(final EffectiveHttpRequestTo effectiveRequest) {
                        result.setDynamicRequest(effectiveRequest.isDynamic());
                        result.setLocalRequest(effectiveRequest.isLocal());
                        if (stream.isClosed()) {
                            deferred.resolve();
                            return Promises.of();
                        }

                        return httpClient.send(effectiveRequest.getHttpRequestTo()).then(new PromiseHandler<HttpClientResultTo>() {

                            @Override
                            public Object on(final HttpClientResultTo httpResult) {
                                onRequestComplete.consume(httpResult);
                                if (stream.isClosed()) {
                                    deferred.resolve();
                                    return Promises.of();
                                }

                                return callback.onRequestComplete(result.getHttpRequest().getId(), result.getHttpRequest(), httpResult)
                                        .then(new PromiseHandler<Void>() {

                                            @Override
                                            public Object on(Void whatever) {

                                                if (httpResult != null) {
                                                    result.setHttpResult(httpResult);

                                                    if (atLeastOneEnabledAssertion(effectiveRequest.getHttpRequestTo().getAssertions())) {
                                                        return evaluateAssertions(effectiveRequest.getHttpRequestTo(), httpResult, deferred, result);
                                                    }

                                                    result.setState(State.Ok);
                                                    if (httpResult.getResponse() != null
                                                            && httpResult.getResponse().getStatus() != null
                                                            && httpResult.getResponse().getStatus().getCode() != null) {

                                                        boolean isValid = Sequence.of(acceptedCodeRanges)
                                                                .some(new Predicate<HttpStatusCodeRange>() {
                                                                    @Override
                                                                    public boolean test(HttpStatusCodeRange acceptableStatusCode) {
                                                                        return acceptableStatusCode.contains(httpResult.getResponse().getStatus().getCode());
                                                                    }
                                                                });

                                                        if (ZERO.contains(httpResult.getResponse().getStatus().getCode())) {
                                                            result.setState(State.Error);
                                                            result.setMessage("The server may be unavailable");
                                                        } else if (!isValid) {
                                                            result.setState(State.Failure);
                                                            result.setMessage("Response status code is " + httpResult.getResponse().getStatus().getCode());
                                                        }
                                                    }
                                                } else {
                                                    result.setMessage("Unknown Error: null response");
                                                    result.setState(State.Error);
                                                }

                                                result.setHttpResult(httpResult);
                                                deferred.resolve();
                                                return null;

                                            }
                                        }).onError(deferred.rejector());

                            }
                        }, new PromiseHandler<Throwable>() {

                            @Override
                            public Object on(Throwable error) {
                                if (stream.isClosed()) {
                                    return null;
                                }

                                result.setThrowable(error);
                                result.setMessage(error.getMessage());
                                result.setState(State.Error);
                                deferred.resolve();
                                return null;
                            }
                        });
                    }

                }, new PromiseHandler<Throwable>() {

                    @Override
                    public Object on(Throwable error) {
                        result.setThrowable(error);
                        result.setMessage(error.getMessage());
                        result.setState(State.Error);
                        deferred.resolve();
                        return null;
                    }
                });
            }
        });
        return deferred.promise();
    }

    private boolean atLeastOneEnabledAssertion(List<AssertionTo> assertions) {
        return Sequence.of(assertions).some(new Predicate<AssertionTo>() {
            @Override
            public boolean test(AssertionTo assertion) {
                return EntityUtils.isEnabled(assertion);
            }
        });
    }

    private Object evaluateAssertions(HttpRequestTo effective, HttpClientResultTo httpResult, final Deferred<Void> deferred, final TestResult result) {
        final List<AssertionResultTo> results = new ArrayList<>();
        for (AssertionTo assertion : effective.getAssertions()) {
            if (assertion != null) {
                results.add(new AssertionResultTo(assertion));
            }
        }

        assertionEngine.evaluateAndCheck(results, httpResult, contextsDao, repositoryDao).then(new PromiseHandler<Void>() {

            @Override
            public Object on(Void whatever) {
                if (stream.isClosed()) {
                    deferred.resolve();
                    return Promises.of();
                }

                for (AssertionResultTo ar : results) {
                    if (!assertionStatesThatAlterRequestState.contains(ar.getState())) {
                        continue;
                    }
                    if (result.getState() == State.Failure
                            && ar.getState() == AssertionResultTo.State.Error) {
                        // error status supersedes failure.
                        setStateAndMessageFromAssertion(result, ar);
                    } else if (result.getState() == State.None
                            || result.getState() == State.InProgress
                            || result.getState() == State.Ok) {
                        setStateAndMessageFromAssertion(result, ar);
                    }
                }
                result.setAssertionResults(results);
                deferred.resolve();
                return null;
            }

        }).onError(deferred.rejector());

        return null;
    }

    private void setStateAndMessageFromAssertion(TestResult result, AssertionResultTo ar) {
        switch (ar.getState()) {
            case Error:
                completeResultMessage(result, ar, AssertionResultTo.State.Error);
                result.setState(State.Error);
                break;
            case Failure:
                completeResultMessage(result, ar, AssertionResultTo.State.Failure);
                result.setState(State.Failure);
                break;
            case Ok:
                result.setState(State.Ok);
                break;
        }
    }

    private void completeResultMessage(TestResult result, AssertionResultTo ar, AssertionResultTo.State state) {
        for (AssertionResultTo.AssertionResultMessage message : ar.getMessages(state)) {
            result.setMessage(message.message);
        }
    }

    @Override
    public Promise<Void> cancel() {
        if (stream != null) {
            stream.cancel();
        }
        if (httpClient.isPending()) {
            httpClient.cancel();
        }
        return Promises.of();
    }

    @Override
    public TestRunner setContextsDao(ContextsDao contextsDao) {
        this.contextsDao = contextsDao;
        return this;
    }

    @Override
    public TestRunner setRepositoryDao(RepositoryDao repositoryDao) {
        this.repositoryDao = repositoryDao;
        return this;
    }

    private static class TestCaseResultStream extends WritableStreamWithResult<TestResult, TestCaseResult> {
        final TestCaseResult testCaseResult;

        TestCaseResultStream(TestCaseResult testCaseResult) {
            this.testCaseResult = testCaseResult;
        }

        void cancel() {
            closeIfStillOpened(testCaseResult);
        }

        @Override
        public AsyncStreamWithResult<TestResult, TestCaseResult> closeIfStillOpened(TestCaseResult result) {
            computeFinalScenarioState(result);
            return super.closeIfStillOpened(result);
        }

        @Override
        public void closeIfStillOpened(Throwable throwable) {
            computeFinalScenarioState(testCaseResult);
            super.closeIfStillOpened(throwable);
        }

        private void computeFinalScenarioState(TestCaseResult result) {
            Map<TestResult.State, List<TestResult>> resultsByState =
                    Sequence.of(result.getTestResults())
                            .groupBy(new Function<TestResult, TestResult.State>() {
                                @Override
                                public State apply(TestResult testResult) {
                                    return testResult.getState();
                                }
                            });

            Sequence.of(resultsByState.get(State.InProgress))
                    .each(new BiConsumer<TestResult, Integer>() {
                        @Override
                        public void consume(TestResult testResult, Integer index) {
                            testResult.setState(State.Skipped);
                            accept(testResult);
                        }
                    });

            result.setState(State.Ok);

            if (resultsByState.containsKey(State.Error)) {
                result.setState(State.Error);
            } else if (resultsByState.containsKey(State.Failure)) {
                result.setState(State.Failure);
            }
        }
    }

}
