package com.restlet.client.utils;

import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import static java.util.Arrays.asList;

/**
 * Sequence is a lazy structure which contains multiple elements similiar to the java.util.Stream
 * or Guava's FluentIterable.
 *
 * @param <T>
 */
public class Sequence<T> implements Iterable<T> {

    protected final Iterable<T> baseIterable;

    public Sequence(final Iterable<T> baseIterable) {
        this.baseIterable = Objects.requireNonNull(baseIterable);
    }

    /**
     * Creates a new sequence from supplied Iterable
     *
     * @param iterable source iterable
     * @param <R>      type of elements in iterable
     * @return sequence based on provided iterable.
     * @deprecated use {@link Sequence#of} instead
     */
    @Deprecated
    public static <R> Sequence<R> from(Iterable<R> iterable) {
        return ofNullable(iterable);
    }

    /**
     * @see com.restlet.client.utils.Sequence#from
     */
    public static <R> Sequence<R> of(Iterable<R> iterable) {
        return ofNullable(iterable);
    }

    /**
     * @see com.restlet.client.utils.Sequence#from
     */
    @SafeVarargs
    public static <R> Sequence<R> of(R... iterable) {
        return ofNullable(asList(iterable));
    }

    private static <R> Sequence<R> ofNullable(@Nullable Iterable<R> iterable) {
        return new Sequence<>(iterable == null
                ? Collections.<R>emptyList()
                : iterable);
    }

    /**
     * Applies function to all elements in source sequence.
     * <p>
     * This operator is non-terminating, e.g. actual mapping of elements will happen only once a terminal operator is called
     *
     * @param mapper mapper function
     * @param <R>    type of elements in returned sequence
     * @return sequence which will return mapped elements
     */
    public <R> Sequence<R> map(final Function<T, R> mapper) {
        return new Sequence(this) {
            @Override
            public Iterator iterator() {
                final Iterator iterator = this.baseIterable.iterator();
                return new Iterator() {
                    @Override
                    public boolean hasNext() {
                        return iterator.hasNext();
                    }

                    @Override
                    public R next() {
                        return mapper.apply((T) iterator.next());
                    }

                    @Override
                    public void remove() {
                        throw new UnsupportedOperationException("remove is not supported");
                    }
                };
            }
        };
    }

    /**
     * Returns a Sequence consisting of the results of replacing each element of
     * this stream with the contents of a mapped stream produced by applying
     * the provided mapping function to each element.
     * <p>
     * This operator is non-terminating, e.g. actual mapping of elements will happen only once a terminal operator is called
     *
     * @param mapper function which maps each element to another iterable
     * @param <R>    type of values for result Sequence
     * @return new Sequence
     */
    public <R> Sequence<R> flatMap(final Function<T, Iterable<R>> mapper) {
        return new Sequence(this) {
            @Override
            public Iterator iterator() {
                return new FlatteningIterator<>(this.baseIterable.iterator(), mapper);
            }
        };
    }

    /**
     * Calls consumer function, with each value in this Sequence
     *
     * @param consumer consumer
     */
    public void each(BiConsumer<T, Integer> consumer) {
        int i = 0;
        for (T t : this) {
            consumer.consume(t, i++);
        }
    }

    public <K> Map<K, List<T>> groupBy(Function<T, K> keyProvider) {
        final Map<K, List<T>> groupedMap = new HashMap<>();
        for (T t : this) {
            K key = keyProvider.apply(t);
            List<T> value = groupedMap.get(key);
            if (value == null) {
                value = new ArrayList<>();
                groupedMap.put(key, value);
            }
            value.add(t);
        }
        return groupedMap;
    }

    public boolean some(Predicate<T> predicate) {
        return this.filter(predicate).count() > 0;
    }

    public boolean has(final T expectedValue) {
        return this.fastReturn(new Predicate<T>() {
            @Override
            public boolean test(final T actualValue) {
                return Objects.equals(actualValue, expectedValue);
            }
        });
    }

    private boolean fastReturn(final Predicate<T> returnOn) {
        for (T element : baseIterable) {
            if (returnOn.test(element)) {
                return true;
            }
        }
        return false;
    }

    public boolean none(Predicate<T> predicate) {
        return !some(predicate);
    }

    public boolean some() {
        return count() > 0;
    }

    public boolean all(Predicate<T> predicate) {
        return this.filter(predicate).count() == this.count();
    }

    public List<T> toList() {
        List<T> arrayList = new ArrayList<>();
        for (T t : this) {
            arrayList.add(t);
        }
        return arrayList;
    }

    public Set<T> toSet() {
        Set<T> set = new HashSet<>();
        for (T t : this) {
            set.add(t);
        }
        return set;
    }

    public int count() {
        int count = 0;
        for (T t : this) {
            count++;
        }
        return count;
    }

    /**
     * First value in the Sequence.
     *
     * @return
     */
    public Maybe<T> first() {
        Iterator<T> iterator = iterator();
        if (iterator.hasNext()) {
            return Maybe.ofNullable(iterator.next());
        }
        return Maybe.empty();
    }

    public <K, V> Map<K, V> toMap(Function<T, K> key, Function<T, V> value) {
        Map<K, V> returnValue = new HashMap<>();
        for (T t : this) {
            returnValue.put(key.apply(t), value.apply(t));
        }
        return returnValue;
    }

    public <K, V> Map<K, List<V>> toEmptyMultiValueMap(Function<T, K> keyGetter) {
        final Map<K, List<V>> resultMap = new HashMap<>();
        for (T t : toSet()) {
            K key = keyGetter.apply(t);
            resultMap.put(key, new ArrayList<V>());
        }

        return resultMap;
    }

    public <K, V> Map<K, List<V>> toMultiValueMap(Function<T, K> keyGetter, Function<T, V> valueGetter) {
        Map<K, List<V>> resultMap = new HashMap<>();
        for (T t : this) {
            K key = keyGetter.apply(t);
            V value = valueGetter.apply(t);
            if (!resultMap.containsKey(key)) {
                resultMap.put(key, new ArrayList<V>());
            }
            resultMap.get(key).add(value);
        }
        return resultMap;
    }

    @Override
    public Iterator<T> iterator() {
        return this.baseIterable.iterator();
    }

    public Sequence<T> omit(final Predicate<T> predicate) {
        final Sequence<T> originalIterator = this;
        return new Sequence<T>(this) {
            @Override
            public Iterator<T> iterator() {
                return new FilteringIterator<>(predicate, originalIterator.iterator(), FilteringOption.INVERTED);
            }
        };
    }

    public Sequence<T> filter(final Predicate<T> predicate) {
        final Sequence<T> originalIterator = this;
        return new Sequence<T>(this) {
            @Override
            public Iterator<T> iterator() {
                return new FilteringIterator<>(predicate, originalIterator.iterator(), FilteringOption.NON_INVERTED);
            }
        };
    }

    public Sequence<T> omitNulls() {
        return filter(new Predicate<T>() {
            @Override
            public boolean test(final T t) {
                return t != null;
            }
        });
    }

    public Sequence<T> omitNonNulls() {
        return filter(new Predicate<T>() {
            @Override
            public boolean test(final T t) {
                return t == null;
            }
        });
    }

    <R> Sequence<R> filterType(final Class<R> expectedType) {
        return filter(new Predicate<T>() {
            @Override
            public boolean test(final T value) {
                return value != null && value.getClass() == expectedType;
            }
        })
                .map(new Function<T, R>() {
                    @Override
                    public R apply(final T input) {
                        return (R) input;
                    }
                });
    }

    public <R> R reduce(BiFunction<R, T, R> reducer, R seed) {
        R nextSeed = seed;
        for (T t : this) {
            nextSeed = reducer.apply(nextSeed, t);
        }
        return nextSeed;
    }

    public Sequence<T> removeDuplicates() {
        final Set<T> uniqueElements = new LinkedHashSet<>();
        return filter(new Predicate<T>() {
            @Override
            public boolean test(final T element) {
                return uniqueElements.add(element);
            }
        });
    }

    public SequenceTuple<T> splitOnPredicate(final Predicate<T> predicate) {
        return new SequenceTuple<>(this.filter(predicate), this.omit(predicate));
    }

    public Sequence<T> reverse() {
        final ArrayList<T> reversed = new ArrayList<>();

        for (T element : baseIterable) {
            reversed.add(0, element);
        }
        return Sequence.ofNullable(reversed);
    }

    public Sequence<T> takeLeft(final int maxIndex) {
        boolean isNegative = maxIndex < 0;
        if (isNegative) {
            throw new IllegalArgumentException("Expected positive integer");
        }
        final ArrayList<T> newIterable = new ArrayList<>();

        each(new BiConsumer<T, Integer>() {
            @Override
            public void consume(T element, Integer index) {
                if (newIterable.size() < maxIndex) {
                    newIterable.add(element);
                }
            }
        });

        return Sequence.ofNullable(newIterable);
    }

    public Sequence<T> takeRight(final int maxIndex) {
        boolean isNegative = maxIndex < 0;
        if (isNegative) {
            throw new IllegalArgumentException("Expected positive integer");
        }
        return reverse().takeLeft(maxIndex);
    }

    public Sequence<T> dropLeft(final int beginningIndex) {
        boolean isNegative = beginningIndex < 0;
        if (isNegative) {
            throw new IllegalArgumentException("Expected positive integer");
        }
        final ArrayList<T> newIterable = new ArrayList<>();

        each(new BiConsumer<T, Integer>() {
            @Override
            public void consume(T element, Integer index) {
                if (index >= beginningIndex) {
                    newIterable.add(element);
                }
            }
        });

        return Sequence.ofNullable(newIterable);
    }

    public Sequence<T> dropRight(final int beginningIndex) {
        boolean isNegative = beginningIndex < 0;
        if (isNegative) {
            throw new IllegalArgumentException("Expected positive integer");
        }
        return reverse().dropLeft(beginningIndex);
    }

    protected static class FlatteningIterator<R, T> implements Iterator<R> {
        final Iterator<T> sourceIterable;
        final Function<T, Iterable<R>> flatFunction;
        Iterator<R> currentIterator;

        public FlatteningIterator(final Iterator<T> sourceIterable, final Function<T, Iterable<R>> flatFunction) {
            this.sourceIterable = sourceIterable;
            this.flatFunction = flatFunction;
        }

        private void trySetupNextIterator() {
            if (currentIterator == null || !currentIterator.hasNext()) {
                while (sourceIterable.hasNext()) {
                    Iterable<R> result = flatFunction.apply(sourceIterable.next());
                    if (result == null) {
                        continue;
                    }
                    currentIterator = result.iterator();
                    if (currentIterator.hasNext()) {
                        break;
                    }
                }
            }
        }

        @Override
        public boolean hasNext() {
            trySetupNextIterator();
            return currentIterator != null && currentIterator.hasNext();
        }

        @Override
        public R next() {
            trySetupNextIterator();
            if (currentIterator == null) {
                throw new IllegalStateException();
            }
            return currentIterator.next();
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

    private enum FilteringOption {
        INVERTED, NON_INVERTED
    }

    private static class FilteringIterator<T> implements Iterator<T> {
        final Predicate<T> filter;
        final Iterator<T> baseIterator;
        private final boolean invertFilter;
        T nextValue;
        boolean hasNextValue = false;

        FilteringIterator(final Predicate<T> filter, final Iterator<T> baseIterator, final FilteringOption filteringOption) {
            this.filter = filter;
            this.baseIterator = baseIterator;
            this.invertFilter = filteringOption == FilteringOption.INVERTED;
        }

        boolean shouldReturnNextValue(T nextValue) {
            if (invertFilter) {
                return !filter.test(nextValue);
            }

            return filter.test(nextValue);
        }

        void ensureNextValue() {
            if (!hasNextValue) {
                while (baseIterator.hasNext()) {
                    T next = baseIterator.next();
                    if (shouldReturnNextValue(next)) {
                        nextValue = next;
                        hasNextValue = true;
                        break;
                    }
                }
            }
        }

        @Override
        public boolean hasNext() {
            ensureNextValue();
            return hasNextValue;
        }

        @Override
        public T next() {
            ensureNextValue();
            if (!hasNextValue) {
                throw new IllegalStateException("Iterator is over");
            }
            T valueToReturn = nextValue;
            nextValue = null;
            hasNextValue = false;
            return valueToReturn;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
}
