package com.restlet.client.utils;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.async.Promises;
import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Executor;
import com.restlet.client.function.Supplier;
import com.restlet.client.log.LogService;

import java.util.LinkedList;
import java.util.Queue;

import static com.restlet.client.utils.StringUtils.format;

public class TaskQueue {

    private final String title;
    private final LogService logService;
    private final Queue<Supplier<Promise<Void>>> queue = new LinkedList<>();

    private TaskQueue(final String title, final LogService logService) {
        this.title = title;
        this.logService = logService;
    }

    public static TaskQueue newTaskQueue(final String title, final LogService logService) {
        return new TaskQueue(title, logService);
    }

    public <T> Promise<T> addToQueue(final String taskTitle, final Supplier<Promise<T>> newPromiseSupplier) {
        final Deferred<T> deferred = new DeferredImpl<>();

        final Supplier<Promise<Void>> newTask = new Supplier<Promise<Void>>() {
            @Override
            public Promise<Void> get() {
                return Promises
                        .create(new Executor() {
                            @Override
                            public void execute() {
                                logService.debug(TaskQueue.class, format("%s BEGIN: %s", title, taskTitle));
                            }
                        })
                        .flatMap(newPromiseSupplier.get())
                        .doOnResolve(new Consumer<T>() {
                            @Override
                            public void consume(final T whatever) {
                                deferred.resolve(whatever);
                                logService.debug(TaskQueue.class, format("%s END: %s", title, taskTitle));
                            }
                        })
                        .recoverOnReject(new PromiseHandler<Throwable>() {
                            @Override
                            public Object on(@Nullable final Throwable error) {
                                deferred.reject(error);
                                logService.error(format("%s ERROR: %s", title, taskTitle), error);
                                return null;
                            }
                        })
                        .doFinally(new Consumer<Deferred>() {
                            @Override
                            public void consume(final Deferred whatever) {
                                queue.poll();

                                if (!queue.isEmpty()) {
                                    queue.peek().get();
                                }
                            }
                        })
                        .castToVoidPromise();
            }
        };

        if (queue.isEmpty()) {
            queue.add(newTask);
            newTask.get();
        } else {
            queue.add(newTask);
        }

        return deferred.promise();
    }
}
